# Code dedicated to cache various useful information
#
# Copyright 2017 Pierre-Yves David <pierre-yves.david@ens-lyon.org>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

import json
import hashlib

from mercurial import (
    error,
    localrepo,
    node,
    obsolete,
)

from . import (
    exthelper,
    obsdiscovery,
    utility,
)

eh = exthelper.exthelper()

class globaljsoncache(dict):

    _filename = None

    def __init__(self):
        self.dirty = False

    def load(self, repo):
        self._dirty = False
        self.clear()
        self.update(self._load(repo))

    def _load(self, repo):
        data = repo.vfs.tryread(self._filename)
        result = {}
        if data:
            parsed = json.loads(data)
            result.update(self._decode(repo, parsed))
        return result

    def _decode(self, repo, fromdisk):
        return fromdisk

    def _encode(self, repo, fromdisk):
        return fromdisk

    def __setitem__(self, key, value):
        if key not in self:
            self._dirty = True
            super(globaljsoncache, self).__setitem__(key, value)

    def save(self, repo):
        # we do not need cache key since the value are global
        if self._dirty:
            wlock = None
            try:
                wlock = repo.wlock(wait=False)
                # re-read the data in case someone updated it since we read it
                #
                # the value we cache are immutable and global so this will be additive only
                self.update(self._load(repo))
                with repo.vfs(self._filename, 'wb', atomictemp=True) as f:
                    json.dump(self._encode(repo, self), f)
                self._dirty = True
            except (IOError, OSError):
                repo.ui.warn('unwritten\n')
                repo.ui.debug('error writing hidden changesets cache\n')
            except error.LockHeld:
                repo.ui.debug('cannot obtain lock to write hidden changesets cache\n')
            finally:
                if wlock:
                    wlock.release()

class depthcache(globaljsoncache):

    _filename = 'cache/depth.json'

    def _decode(self, repo, fromdisk):
        bin = node.bin
        return dict((bin(h), d) for (h, d) in fromdisk.items())

    def _encode(self, repo, todisk):
        hex = node.hex
        return dict((hex(n), d) for (n, d) in todisk.items())

class stablerangecache(globaljsoncache):

    _filename = 'cache/stablerange.json'

    def _decode(self, repo, fromdisk):
        bin = node.bin
        result = {}
        cache = {}
        stablerange = obsdiscovery._range
        unfi = repo.unfiltered()
        rev = unfi.changelog.rev

        def dec(entry):
            sr = cache.get(entry)
            if sr is None:
                h, i = entry.split(':')
                cache[entry] = sr = stablerange(unfi, rev(bin(h)), int(i))
            return sr
        for sup, sub in fromdisk.items():
            result[dec(sup)] = [dec(s) for s in sub]
        return result

    def _encode(self, repo, todisk):
        result = {}
        for sup, sub in todisk.items():
            result[sup.jsonkey] = [s.jsonkey for s in sub]
        return result

class rangeobshash(object):

    _filename = 'cache/obshashrange.json'

    def __init__(self):
        self._cache = {}
        self._dirty = False

    def __getitem__(self, stablerange):
        value = self._cache.get(stablerange)
        if value is not None:
            return value
        pieces = []
        obshash = None
        nullid = node.nullid
        if len(stablerange) == 1:
            rm = stablerange._repo.obsstore.relevantmarkers
            tmarkers = rm([stablerange.node])
            pieces = []
            for m in tmarkers:
                mbin = obsolete._fm1encodeonemarker(m)
                pieces.append(mbin)
            pieces.sort()
        else:
            for subrange in stablerange.subranges():
                obshash = self[subrange]
                if obshash != nullid:
                    pieces.append(obshash)

        sha = hashlib.sha1()
        # note: if there is only one subrange with actual data, we'll just
        # reuse the same hash.
        if not pieces:
            obshash = node.nullid
        elif len(pieces) != 1 or obshash is None:
            sha = hashlib.sha1()
            for p in pieces:
                sha.update(p)
            obshash = self._cache[stablerange] = sha.digest()
        self._dirty = True
        self._cache[stablerange] = obshash
        return obshash

    def clear(self):
        if self._cache:
            self._dirty = True
        self._cache.clear()

    def save(self, repo):
        if self._dirty:
            data = {}
            unfi = repo.unfiltered()
            hex = node.hex
            data['version'] = 1
            data['obsstore-len'] = len(unfi.obsstore)
            if len(unfi.obsstore):
                mhash = obsolete._fm1encodeonemarker(unfi.obsstore._all[-1])
                data['obsstore-tip'] = node.hex(mhash)
            data['heads'] = [node.hex(h) for h in unfi.changelog.heads()]
            data['obshash'] = [(r.jsonkey, hex(v)) for r, v in self._cache.items()]
            wlock = None
            try:
                wlock = repo.wlock(wait=False)
                # re-read the data in case someone updated it since we read it
                #
                # the value we cache are immutable and global so this will be additive only
                with repo.vfs(self._filename, 'wb', atomictemp=True) as f:
                    json.dump(data, f)
                self._dirty = True
            except (IOError, OSError):
                repo.ui.warn('unwritten\n')
                repo.ui.debug('error writing hidden changesets cache\n')
            except error.LockHeld:
                repo.ui.debug('cannot obtain lock to write hidden changesets cache\n')
            finally:
                if wlock:
                    wlock.release()

    def load(self, repo):
        unfi = repo.unfiltered()
        self._cache.clear()
        self._dirty = False
        content = repo.vfs.tryread(self._filename)
        nm = unfi.changelog.nodemap
        rev = unfi.changelog.rev
        if content:
            data = json.loads(content)
            if data['version'] != 1:
                unfi.ui.debug('invalid obshashrange cache (version)\n')
                return
            obslen = data['obsstore-len']
            if obslen != len(unfi.obsstore):
                unfi.ui.debug('invalid obshashrange cache (obsstore-len)\n')
                return
            if obslen:
                mhash = obsolete._fm1encodeonemarker(unfi.obsstore._all[-1])
                if node.bin(data['obsstore-tip']) != mhash:
                    unfi.ui.debug('invalid obshashrange cache (obsstore-tip)\n')
                    return
            for h in data['heads']:
                n = node.bin(h)
                if nm.get(n) is None:
                    unfi.ui.debug('invalid obshashrange cache (missing-heads)\n')
                    return

            stablerange = obsdiscovery._range
            bin = node.bin
            cache = {}
            for k, v in data['obshash']:
                k = str(k)
                sr = cache.get(k)
                if sr is None:
                    h, i = k.split(':')
                    sr = stablerange(unfi, rev(bin(h)), int(i))
                    cache[str(k)] = sr
                self._cache[sr] = bin(v)

@eh.reposetup
def setupcache(ui, repo):

    class cachedrepo(repo.__class__):

        @localrepo.unfilteredpropertycache
        def depthcache(self):
            dc = depthcache()
            dc.load(self)
            return dc

        @localrepo.unfilteredpropertycache
        def stablerangecache(self):
            src = stablerangecache()
            src.load(self)
            # print 'stablerange: have %d entry cached' % len(src)
            # if src:
            #     print 'stablerange[0]:', list(src)[0]
            return src

        def transaction(self, *args, **kwargs):
            tr = super(cachedrepo, self).transaction(*args, **kwargs)

            def _warmcaches(tr):
                if 'node' in tr.hookargs:
                    revs = repo.revs('%n:%n', node.bin(tr.hookargs['node']),
                                     node.bin(tr.hookargs['node_last']))
                    for r in revs:
                        utility.depth(self, r)
                if 'depthcache' in vars(self):
                    self.depthcache.save(self)
                if 'stablerangecache' in vars(self):
                    self.stablerangecache.save(self)
                if 'rangeobshash' in vars(self.obsstore):
                    self.obsstore.rangeobshash.save(self)

            tr.addpostclose('evoexpt_warmcache', _warmcaches)
            return tr

    repo.__class__ = cachedrepo
